#include <feos_array>

using namespace FeOS;

template <size_t elemSize>
_ArrayBase<elemSize>::_ArrayBase(const void* buf, size_t elems) noexcept
	: _ArrayBase()
{
	if (elems && SetCapacity(elems))
	{
		memcpy(m_buffer, buf, elems*elemSize);
		m_elemCount = elems;
	}
}

template <size_t elemSize>
bool _ArrayBase<elemSize>::SetCapacity(size_t elems) noexcept
{
	if (elems == 0)
	{
		free(m_buffer);
		m_buffer = nullptr;
		m_capacity = 0;
		m_elemCount = 0;
		return true;
	}

	void* newBuf = realloc(m_buffer, elemSize*elems);
	if (newBuf)
	{
		m_buffer = newBuf;
		m_capacity = elems;
		if (m_elemCount > elems)
			m_elemCount = elems;
		return true;
	} else
		return false;
}

template <size_t elemSize>
void* _ArrayBase<elemSize>::AddElem(size_t count) noexcept
{
	if (!count || ((m_elemCount + count) > m_capacity && !SetCapacity(m_capacity + Max(count, 16U))))
		return nullptr;

	void* elem = end();
	m_elemCount += count;
	return elem;
}

template <size_t elemSize>
void* _ArrayBase<elemSize>::AddElemAt(size_t pos, size_t count) noexcept
{
	if (pos == m_elemCount)
		return AddElem(count);

	if (!count || ((m_elemCount + count) > m_capacity && !SetCapacity(m_capacity + Max(count, 16U))))
		return nullptr;

	void* elem = ElemAt(pos);
	memmove((char*)elem + count*elemSize, elem, (m_elemCount - pos) * elemSize);
	m_elemCount += count;
	return elem;
}

template <size_t elemSize>
void _ArrayBase<elemSize>::RemoveRange(size_t begin, size_t end) noexcept
{
	ptrdiff_t diff = (ptrdiff_t)end - (ptrdiff_t)begin;
	if (diff <= 0) return;

	size_t movesize = (m_elemCount - end) * elemSize;
	m_elemCount -= diff;

	if (movesize)
		memcpy(ElemAt(begin), ElemAt(end), movesize);
}

#define IMPL
_ARRAYBASE_EXTERNALS(IMPL, 1);
_ARRAYBASE_EXTERNALS(IMPL, 2);
_ARRAYBASE_EXTERNALS(IMPL, 4);
