#include <feos_deque>

using namespace FeOS;

void* _DequeBase::_insertCommon(bool isBack, size_t elemSize, size_t sliceShift) noexcept
{
	size_t sliceCount = m_slices.Count();
	size_t sliceSize = elemSize << sliceShift;

	if (m_elemCount == 0)
	{
		void* mem1 = nullptr;
		void* mem2 = nullptr;

		if (sliceCount == 0)
		{
			// Prepare the container
			if (!m_slices.EnsureCapacity(16))
				return nullptr;

			mem1 = malloc(sliceSize);
			if (!mem1)
				return nullptr;

			mem2 = malloc(sliceSize);
			if (!mem2)
			{
				free(mem1);
				return nullptr;
			}

			m_slices.Add(mem1);
			m_slices.Add(mem2);

			m_first = (1U<<sliceShift) + (size_t)isBack - 1;
		} else
		{
			// Use the slices in the middle
			size_t p = (sliceCount-2) / 2;
			mem1 = m_slices[p];
			mem2 = m_slices[p+1];

			m_first = ((p+1)<<sliceShift) + (size_t)isBack - 1;
		}

		m_elemCount = 1;
		return isBack ? mem2 : ((char*)mem1 + sliceSize - elemSize);
	}

	size_t sliceElemCount = 1U << sliceShift;

	if (isBack)
	{
		size_t p = m_first + m_elemCount;
		if (p != Capacity(sliceShift))
		{
			m_elemCount ++;
			return _rawElemAt(p, elemSize, sliceShift);
		}

		void* mem = nullptr;

		// Check whether we can reuse an empty beginning slice
		if (sliceElemCount < m_first)
		{
			// Yes!
			mem = m_slices[0];
			m_slices.Remove((size_t)0);
			m_slices.Add(mem);
		} else
		{
			// Need a new slice
			mem = malloc(sliceSize);
			if (!mem) return nullptr;
			if (!m_slices.Add(mem))
			{
				// Rare, although it /could/ happen
				free(mem);
				return nullptr;
			}
		}

		m_elemCount ++;
		return mem;
	} else
	{
		if (m_first != 0)
		{
			m_first --;
			m_elemCount ++;
			return _rawElemAt(m_first, elemSize, sliceShift);
		}

		size_t endp = m_first + m_elemCount;
		size_t lastS = (sliceCount-1) << sliceShift;
		void* mem = nullptr;
		
		// Check whether we can reuse an empty ending slice
		if (lastS > endp)
		{
			// Yes!
			mem = m_slices[sliceCount-1];
			m_slices.Remove(sliceCount-1);
			m_slices.Insert((size_t)0, mem);
		} else
		{
			// Need a new slice
			mem = malloc(sliceSize);
			if (!mem) return nullptr;
			if (!m_slices.Insert((size_t)0, mem))
			{
				// Rare, although it /could/ happen
				free(mem);
				return nullptr;
			}
		}

		m_first += sliceElemCount - 1;
		m_elemCount ++;
		return (char*)mem + sliceSize - elemSize;
	}
}

_DequeBase::~_DequeBase() noexcept
{
	for (auto x : m_slices)
		free(x);
}
