#include <stdio.h>
#include <feos_list>
#include <feos.h>

using namespace FeOS;

auto _ListBase::_newNode(size_t elemSize) noexcept -> node*
{
	if (!m_free)
		return (node*) malloc(sizeof(node) + elemSize);

	// Else get a free one
	node* p = m_free;
	m_free = p->next;
	return p;
}

auto _ListBase::Insert(node* pos, size_t elemSize) noexcept -> node*
{
	node* n = _newNode(elemSize);
	if (!n) return nullptr;

	if (m_elemCount++ == 0) // No elements.
	{
		n->prev = nullptr;
		n->next = nullptr;
		m_first = n;
		m_last = n;
	} else if (pos == nullptr) // Insert at container end.
	{
		n->prev = m_last;
		n->next = nullptr;
		m_last->next = n;
		m_last = n;
	} else if (pos == m_first) // Insert at container start.
	{
		n->prev = nullptr;
		n->next = m_first;
		m_first->prev = n;
		m_first = n;
	} else // Insert in the middle of the container.
	{
		n->prev = pos->prev;
		n->next = pos;
		pos->prev->next = n;
		pos->prev = n;
	}

	return n;
}

void _ListBase::Clear() noexcept
{
	if (!m_elemCount)
		return;

	// Perform a fake free - move all the elements to the free element linked list
	m_elemCount = 0;
	m_last->next = m_free;
	m_free = m_first;
	m_first = nullptr;
	m_last = nullptr;
}

void _ListBase::DeleteCache() noexcept
{
	node* next = nullptr;
	for (node* n = m_free; n != nullptr; n = next)
	{
		next = n->next;
		free(n);
	}
	m_free = nullptr;
}

void _ListBase::Remove(node* pos) noexcept
{
	if (!pos) return; // Nothing to do.
	node* oldPrev = pos->prev;
	node* oldNext = pos->next;
	// Push this node to the free node stash
	pos->next = m_free;
	m_free = pos;

	if (!--m_elemCount) // The only element.
	{
		m_first = nullptr;
		m_last = nullptr;
	} else if (!oldPrev) // First element.
	{
		m_first = oldNext;
		oldNext->prev = nullptr;
	} else if (!oldNext) // Last element.
	{
		m_last = oldPrev;
		oldPrev->next = nullptr;
	} else // Middle element.
	{
		oldPrev->next = oldNext;
		oldNext->prev = oldPrev;
	}
}

void _ListBase::RemoveRange(node* begin, node* end, size_t elemCountHint) noexcept
{
	if (!begin || begin == end) return; // Nothing to do.

	m_elemCount -= elemCountHint;

	if (end == nullptr) // Remove the tail of the list
	{
		m_last->next = m_free;
		m_last = begin->prev;
		if (m_last)
			m_last->next = nullptr;
		else // Remove the whole list
			m_first = nullptr;
		m_free = begin;
	} else
	{
		node* befBegin = begin->prev;
		node* befEnd = end->prev;

		if (befBegin == nullptr) // Remove the head of the list
			m_first = end;
		else // Middle of the list
			befBegin->next = end;

		end->prev = befBegin;
		befEnd->next = m_free;
		m_free = begin;
	}
}
